@@ -7,21 +7,39 @@
 #
 import csv
 import sys
+import signal
 import shutil
+import importlib
 from textwrap import dedent
-from argparse import ArgumentParser, FileType, RawTextHelpFormatter
+from argparse import (
+    ArgumentParser,
+    ArgumentTypeError,
+    FileType,
+    RawTextHelpFormatter
+)
 
+from minet.__version__ import __version__
 from minet.defaults import DEFAULT_THROTTLE
 from minet.cli.defaults import DEFAULT_CONTENT_FOLDER
-from minet.cli.utils import BooleanAction
+from minet.cli.utils import BooleanAction, die
 
-from minet.cli.crowdtangle.constants import CROWDTANGLE_SORT_TYPES, CROWDTANGLE_DEFAULT_RATE_LIMIT
+from minet.cli.crowdtangle.constants import (
+    CROWDTANGLE_SORT_TYPES,
+    CROWDTANGLE_DEFAULT_RATE_LIMIT,
+    CROWDTANGLE_PARTITION_STRATEGIES
+)
 
 SUBPARSERS = {}
 
+# Getting terminal size
 terminal_size = shutil.get_terminal_size()
+
+# Increasing max CSV file limit to avoid pesky issues
 csv.field_size_limit(sys.maxsize)
 
+# Hiding stack traces on ctrl+c
+signal.signal(signal.SIGINT, lambda x, y: sys.exit(1))
+
 
 def custom_formatter(prog):
     return RawTextHelpFormatter(
@@ -43,17 +61,60 @@ def omit(d, key_to_omit):
     return nd
 
 
+def partition_strategy_type(string):
+    if string in CROWDTANGLE_PARTITION_STRATEGIES:
+        return string
+
+    try:
+        return int(string)
+    except ValueError:
+        choices = ' or '.join(CROWDTANGLE_PARTITION_STRATEGIES)
+
+        raise ArgumentTypeError('partition strategy should either be %s, or an number of posts.' % choices)
+
+
+def get_subparser(o, keys):
+    parser = None
+
+    for key in keys:
+        item = o.get(key)
+
+        if item is None:
+            return None
+
+        parser = item['parser']
+
+        if 'subparsers' in item:
+            o = item['subparsers']
+        else:
+            break
+
+    return parser
+
+
+def check_dragnet():
+    try:
+        import dragnet
+    except:
+        die([
+            'The `dragnet` library is not installed. The `extract` command won\'t work.',
+            'To install it correctly, run the following commands in order:',
+            '',
+            '  pip install lxml numpy Cython',
+            '  pip install dragnet'
+        ])
+
+
 # Defining the list of CLI commands
-# TODO: generate docs from this data
 COMMANDS = {
 
     # Fetch action subparser
     # --------------------------------------------------------------------------
     'fetch': {
+        'package': 'minet.cli.fetch',
+        'action': 'fetch_action',
+        'title': 'Minet Fetch Command',
         'description': '''
-            Minet Fetch Command
-            ===================
-
             Use multiple threads to fetch batches of urls from a CSV file. The
             command outputs a CSV report with additional metadata about the
             HTTP calls and will generally write the retrieved files in a folder
@@ -84,9 +145,10 @@ COMMANDS = {
                 'nargs': '?'
             },
             {
-                'flag': '--contents-in-report',
+                'flags': ['--contents-in-report', '--no-contents-in-report'],
                 'help': 'Whether to include retrieved contents, e.g. html, directly in the report\nand avoid writing them in a separate folder. This requires to standardize\nencoding and won\'t work on binary formats.',
-                'action': 'store_true'
+                'dest': 'contents_in_report',
+                'action': BooleanAction
             },
             {
                 'flags': ['-d', '--output-dir'],
@@ -103,8 +165,8 @@ COMMANDS = {
             },
             {
                 'flags': ['-g', '--grab-cookies'],
-                'help': 'Whether to attempt to grab cookies from your computer\'s chrome browser.',
-                'action': 'store_true'
+                'help': 'Whether to attempt to grab cookies from your computer\'s browser.',
+                'choices': ['firefox', 'chrome']
             },
             {
                 'flags': ['-H', '--header'],
@@ -145,17 +207,24 @@ COMMANDS = {
             {
                 'flag': '--url-template',
                 'help': 'A template for the urls to fetch. Handy e.g. if you need to build urls from ids etc.'
+            },
+            {
+                'flags': ['-X', '--request'],
+                'help': 'The http method to use. Will default to GET.',
+                'dest': 'method',
+                'default': 'GET'
             }
         ]
     },
 
     # Crowdtangle action subparser
     # --------------------------------------------------------------------------
-    'ct': {
+    'crowdtangle': {
+        'package': 'minet.cli.crowdtangle',
+        'action': 'crowdtangle_action',
+        'aliases': ['ct'],
+        'title': 'Minet Crowdtangle Command',
         'description': '''
-            Minet CrowdTangle Command
-            =========================
-
             Gather data from the CrowdTangle APIs easily and efficiently.
         ''',
         'subparsers': {
@@ -180,10 +249,8 @@ COMMANDS = {
             ],
             'commands': {
                 'leaderboard': {
+                    'title': 'Minet CrowdTangle Leaderboard Command',
                     'description': '''
-                        Minet CrowdTangle Leaderboard Command
-                        =====================================
-
                         Gather information and aggregated stats about pages and groups of
                         the designated dashboard (indicated by a given token).
                     ''',
@@ -219,10 +286,8 @@ COMMANDS = {
                     ]
                 },
                 'lists': {
+                    'title': 'Minet CrowdTangle Lists Command',
                     'description': '''
-                        Minet CrowdTangle Lists Command
-                        ===============================
-
                         Retrieve the lists from a CrowdTangle dashboard (indicated by a
                         given token).
                     ''',
@@ -234,10 +299,8 @@ COMMANDS = {
                     '''
                 },
                 'posts': {
+                    'title': 'Minet CrowdTangle Posts Command',
                     'description': '''
-                        Minet CrowdTangle Posts Command
-                        ===============================
-
                         Gather post data from the designated dashboard (indicated by
                         a given token).
                     ''',
@@ -271,6 +334,16 @@ COMMANDS = {
                             'flag': '--list-ids',
                             'help': 'Ids of the lists from which to retrieve posts, separated by commas.'
                         },
+                        {
+                            'flag': '--partition-strategy',
+                            'help': 'Query partition strategy to use to overcome the API search result limits. Should either be `day` or a number of posts.',
+                            'type': partition_strategy_type
+                        },
+                        {
+                            'flag': '--resume',
+                            'help': 'Whether to resume an interrupted collection. Requires -o/--output & --sort-by date',
+                            'action': 'store_true'
+                        },
                         {
                             'flag': '--sort-by',
                             'help': 'The order in which to retrieve posts. Defaults to `date`.',
@@ -289,10 +362,8 @@ COMMANDS = {
                     ]
                 },
                 'search': {
+                    'title': 'Minet CrowdTangle Search Command',
                     'description': '''
-                        Minet CrowdTangle Search Command
-                        ================================
-
                         Search posts on the whole CrowdTangle platform.
                     ''',
                     'epilog': '''
@@ -321,6 +392,11 @@ COMMANDS = {
                             'help': 'Maximum number of posts to retrieve. Will fetch every post by default.',
                             'type': int
                         },
+                        {
+                            'flag': '--not-in-title',
+                            'help': 'Whether to search terms in account titles also.',
+                            'action': 'store_true'
+                        },
                         {
                             'flag': '--offset',
                             'help': 'Count offset.',
@@ -328,8 +404,8 @@ COMMANDS = {
                         },
                         {
                             'flag': '--partition-strategy',
-                            'help': 'Query partition strategy to use to overcome the API search result limits.',
-                            'choices': ['day']
+                            'help': 'Query partition strategy to use to overcome the API search result limits. Should either be `day` or a number of posts.',
+                            'type': partition_strategy_type
                         },
                         {
                             'flags': ['-p', '--platforms'],
@@ -363,10 +439,10 @@ COMMANDS = {
     # Extract action subparser
     # -------------------------------------------------------------------------
     'extract': {
+        'package': 'minet.cli.extract',
+        'action': 'extract_action',
+        'title': 'Minet Extract Command',
         'description': '''
-            Minet Extract Command
-            =====================
-
             Use multiple processes to extract raw text from a batch of HTML files.
             This command can either work on a `minet fetch` report or on a bunch
             of files. It will output an augmented report with the extracted text.
@@ -383,6 +459,7 @@ COMMANDS = {
             . Extracting raw text from a bunch of files:
                 `minet extract --glob "./content/*.html" > extracted.csv`
         ''',
+        'before': check_dragnet,
         'arguments': [
             {
                 'name': 'report',
@@ -425,11 +502,12 @@ COMMANDS = {
 
     # Facebook actions subparser
     # -------------------------------------------------------------------------
-    'fb': {
+    'facebook': {
+        'package': 'minet.cli.facebook',
+        'action': 'facebook_action',
+        'aliases': ['fb'],
+        'title': 'Minet Facebook Command',
         'description': '''
-            Minet Facebook Command
-            ======================
-
             Collects data from Facebook.
         ''',
         'subparsers': {
@@ -438,10 +516,8 @@ COMMANDS = {
             'dest': 'fb_action',
             'commands': {
                 'comments': {
+                    'title': 'Minet Facebook Comments Command',
                     'description': '''
-                        Minet Facebook Comments Command
-                        ===============================
-
                         Scrape series of comments on Facebook.
                     ''',
                     'epilog': '''
@@ -473,10 +549,10 @@ COMMANDS = {
     # Scrape action subparser
     # -------------------------------------------------------------------------
     'scrape': {
+        'package': 'minet.cli.scrape',
+        'action': 'scrape_action',
+        'title': 'Minet Scrape Command',
         'description': '''
-            Minet Scrape Command
-            ====================
-
             Use multiple processes to scrape data from a batch of HTML files.
             This command can either work on a `minet fetch` report or on a bunch
             of files. It will output the scraped items.
@@ -488,7 +564,10 @@ COMMANDS = {
                 `minet scrape scraper.json report.csv > scraped.csv`
 
             . Working on a report from stdin:
-                `minet fetch url_column file.csv | minet fetch scraper.json > scraped.csv`
+                `minet fetch url_column file.csv | minet scrape scraper.json > scraped.csv`
+
+            . Scraping a single page from the web:
+                `minet fetch https://news.ycombinator.com/ | minet scrape scraper.json > scraped.csv`
 
             . Scraping items from a bunch of files:
                 `minet scrape scraper.json --glob "./content/*.html" > scraped.csv`
@@ -506,6 +585,16 @@ COMMANDS = {
                 'default': sys.stdin,
                 'nargs': '?'
             },
+            {
+                'flags': ['-f', '--format'],
+                'help': 'Output format.',
+                'choices': ['csv', 'jsonl'],
+                'default': 'csv'
+            },
+            {
+                'flags': ['-g', '--glob'],
+                'help': 'Whether to scrape a bunch of html files on disk matched by a glob pattern rather than sourcing them from a CSV report.'
+            },
             {
                 'flags': ['-i', '--input-directory'],
                 'help': 'Directory where the HTML files are stored. Defaults to "%s".' % DEFAULT_CONTENT_FOLDER,
@@ -541,10 +630,23 @@ def add_arguments(subparser, arguments):
             subparser.add_argument(*argument['flags'], **omit(argument, 'flags'))
 
 
-def main():
+def build_description(command):
+    description = command['title'] + '\n' + ('=' * len(command['title']))
+
+    description += '\n\n' + dedent(command.get('description', ''))
+
+    return description
+
+
+def build_parser(commands):
 
     # Building the argument parser
     parser = ArgumentParser(prog='minet')
+
+    parser.add_argument('--version', action='version', version='minet %s' % __version__)
+
+    subparser_index = {}
+
     subparsers = parser.add_subparsers(
         help='Action to execute',
         title='actions',
@@ -554,11 +656,18 @@ def main():
     for name, command in COMMANDS.items():
         subparser = subparsers.add_parser(
             name,
-            description=dedent(command['description']),
+            description=build_description(command),
             epilog=dedent(command.get('epilog', '')),
-            formatter_class=custom_formatter
+            formatter_class=custom_formatter,
+            aliases=command.get('aliases', [])
         )
 
+        to_index = {
+            'parser': subparser,
+            'command': command,
+            'subparsers': {}
+        }
+
         if 'arguments' in command:
             add_arguments(subparser, command['arguments'])
 
@@ -578,7 +687,7 @@ def main():
             for subname, subcommand in command['subparsers']['commands'].items():
                 subsubparser = subparser_subparsers.add_parser(
                     subname,
-                    description=dedent(subcommand['description']),
+                    description=build_description(subcommand),
                     epilog=dedent(subcommand.get('epilog', '')),
                     formatter_class=custom_formatter
                 )
@@ -589,54 +698,52 @@ def main():
                 if 'arguments' in subcommand:
                     add_arguments(subsubparser, subcommand['arguments'])
 
-        SUBPARSERS[name] = subparser
+                to_index['subparsers'][subname] = {
+                    'parser': subsubparser
+                }
+
+        if 'aliases' in command:
+            for alias in command['aliases']:
+                subparser_index[alias] = to_index
+
+        subparser_index[name] = to_index
 
     # Help subparser
-    help_suparser = subparsers.add_parser('help')
-    help_suparser.add_argument('subcommand', help='Name of the subcommand', nargs='?')
-    SUBPARSERS['help'] = help_suparser
+    help_subparser = subparsers.add_parser('help')
+    help_subparser.add_argument('subcommand', help='Name of the subcommand', nargs='*')
+
+    return parser, subparser_index
+
+
+def main():
+
+    # Building parser
+    parser, subparser_index = build_parser(COMMANDS)
 
     # Parsing arguments and triggering commands
     args = parser.parse_args()
 
-    if args.action == 'ct':
-        from minet.cli.crowdtangle import crowdtangle_action
-        crowdtangle_action(args)
+    action = subparser_index.get(args.action)
 
-    elif args.action == 'extract':
-        try:
-            import dragnet
-        except:
-            print('The `dragnet` library is not installed. The `extract` command won\'t work.')
-            print('To install it correctly, run the following commands in order:')
-            print()
-            print('  pip install lxml numpy Cython')
-            print('  pip install dragnet')
-            sys.exit(1)
+    if action is not None:
 
-        from minet.cli.extract import extract_action
-        extract_action(args)
+        # Need to check something?
+        if 'before' in action['command']:
+            action['command']['before']()
 
-    elif args.action == 'fb':
-        from minet.cli.facebook import facebook_action
-        facebook_action(args)
+        # Lazy loading module for faster startup
+        m = importlib.import_module(action['command']['package'])
+        fn = getattr(m, action['command']['action'])
 
-    elif args.action == 'fetch':
-        from minet.cli.fetch import fetch_action
-        fetch_action(args)
+        fn(args)
 
     elif args.action == 'help':
-        # TODO: handle sub commands?
-        target_subparser = SUBPARSERS.get(args.subcommand)
+        target = get_subparser(subparser_index, args.subcommand)
 
-        if target_subparser is None:
-            parser.print_help()
+        if target is None:
+            die('Unknow command "%s"' % ' '.join(args.subcommand))
         else:
-            target_subparser.print_help()
-
-    elif args.action == 'scrape':
-        from minet.cli.scrape import scrape_action
-        scrape_action(args)
+            target.print_help()
 
     else:
         parser.print_help()
